#include <algorithm>
#include <cmath>
#include <cstdlib>
#include <limits>
#include <signal.h>
#include <msp/fs/stat.h>
#include <msp/fs/utils.h>
#include <msp/graphics/display.h>
#include <msp/graphics/window.h>
#include <msp/gl/blend.h>
#include <msp/gl/framebuffer.h>
#include <msp/gl/matrix.h>
#include <msp/gl/misc.h>
#include <msp/gl/tests.h>
#include <msp/gltk/floatingarrangement.h>
#include <msp/io/print.h>
#include <msp/strings/format.h>
#include <msp/time/units.h>
#include <msp/time/utils.h>
#include "libr2c2/driver.h"
#include "libr2c2/trackcircuit.h"
#include "libr2c2/tracktype.h"
#include "3d/allocation.h"
#include "3d/path.h"
#include "3d/track.h"
#include "3d/trackcircuit.h"
#include "3d/vehicle.h"
#include "engineer.h"
#include "mainwindow.h"
#include "newtraindialog.h"
#include "traindialog.h"
#include "trainview.h"

using namespace std;
using namespace R2C2;
using namespace Msp;

Engineer::Engineer(int argc, char **argv):
	options(argc, argv),
	window(options.screen_w, options.screen_h, options.fullscreen),
	keyboard(window),
	mouse(window),
	ui_res("data/r2c2.res"),
	import_active(false),
	layout(catalogue, (options.driver.empty() ? 0 : Driver::create(options.driver))),
	layout_3d(layout),
	server(0),
	main_view(layout_3d, window.get_width(), window.get_height()),
	emergency_blink_state(0)
{
	// Setup GUI
	window.set_title("Railroad Engineer");
	window.signal_close.connect(sigc::bind(sigc::mem_fun(this, &Engineer::exit), 0));

	root = new GLtk::Root(ui_res, &window, &keyboard, &mouse);
	GLtk::Layout *root_layout = new GLtk::Layout;
	root->set_layout(root_layout);
	root_layout->set_margin(GLtk::Sides());
	root_arrangement = new GLtk::FloatingArrangement(*root_layout);
	mouse.signal_button_press.connect(sigc::bind_return(sigc::mem_fun(this, &Engineer::button_press), false));
	mouse.signal_axis_motion.connect(sigc::bind_return(sigc::mem_fun(this, &Engineer::axis_motion), false));
	root->set_visible(true);

	main_wnd = new MainWindow(*this);
	root->add(*main_wnd);
	main_wnd->autosize();
	main_wnd->set_position(0, window.get_height()-main_wnd->get_geometry().h);

	overlay = new Overlay3D(ui_res.get_default_font());

	// Setup railroad control
	catalogue.add_source("data/Märklin/H0");
	DataFile::load(layout, options.layout_fn);

	if(layout.has_driver())
	{
		Driver &driver = layout.get_driver();
		driver.signal_locomotive_detected.connect(sigc::mem_fun(this, &Engineer::locomotive_detected));
		driver.signal_halt.connect(sigc::mem_fun(this, &Engineer::halt_event));
	}
	layout.signal_train_added.connect(sigc::mem_fun(this, &Engineer::train_added));
	layout.signal_emergency.connect(sigc::mem_fun(this, &Engineer::emergency));
	const set<Block *> &blocks = layout.get_all<Block>();
	for(set<Block *>::const_iterator i=blocks.begin(); i!=blocks.end(); ++i)
		if(TrackCircuit *tc = (*i)->get_sensor())
			new TrackCircuit3D(layout_3d, *tc);

	const set<Track *> &tracks = layout.get_all<Track>();
	for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
		if((*i)->get_type().is_turnout())
			new Path3D(layout_3d.get_3d(**i));

	if(FS::exists(options.state_fn))
		DataFile::load(layout, options.state_fn);

	if(options.network)
	{
		server = new Server(layout);
		server->use_event_dispatcher(event_disp);
	}

	// Setup 3D view
	GL::Pipeline &pipeline = main_view.get_pipeline();

	GL::Pipeline::Pass *pass = &pipeline.add_pass("unlit");
	pass->set_depth_test(&GL::DepthTest::lequal());
	pipeline.add_renderable_for_pass(layout_3d.get_path_scene(), "unlit");

	pass = &pipeline.add_pass("overlay");
	pass->set_blend(&GL::Blend::alpha());
	pipeline.add_renderable_for_pass(*overlay, "overlay");

	view_all();

	// Catch various signals so we can stop the trains in case we get terminated
	catch_signal(SIGINT);
	catch_signal(SIGTERM);
	catch_signal(SIGSEGV);
	catch_signal(SIGILL);
	catch_signal(SIGFPE);
	catch_signal(SIGABRT);
}

Engineer::~Engineer()
{
	if(!options.simulate)
	{
		layout.save_dynamic(options.state_fn+".tmp");
		FS::rename(options.state_fn+".tmp", options.state_fn);
	}

	layout.get_driver().halt(true);
	layout.get_driver().flush();

	delete overlay;
	delete root_arrangement;
	delete root;

	delete server;
}

void Engineer::set_status(const string &text)
{
	main_wnd->set_status_text(text);
	status_timeout = Time::now()+10*Time::sec;
}

void Engineer::add_train_view(TrainView &tv)
{
	train_views.push_back(&tv);
}

void Engineer::remove_train_view(TrainView &tv)
{
	train_views.erase(remove(train_views.begin(), train_views.end(), &tv), train_views.end());
}

int Engineer::main()
{
	window.show();

	return Application::main();
}

void Engineer::tick()
{
	window.get_display().tick();

	for(list<Train *>::iterator i=new_trains.begin(); i!=new_trains.end(); ++i)
		process_new_train(**i);
	new_trains.clear();

	layout.tick();
	layout_3d.tick();
	event_disp.tick(Time::zero);

	for(list<TrainView *>::iterator i=train_views.begin(); i!=train_views.end(); ++i)
		(*i)->prepare();

	Time::TimeStamp t = Time::now();
	if(status_timeout && t>status_timeout)
	{
		main_wnd->set_status_text(string());
		status_timeout = Time::TimeStamp();
	}
	if(!emergencies.empty() && t>emergency_blink_timeout)
	{
		emergency_blink_state = (emergency_blink_state+1)%2;
		GL::Color color(1.0f/(1+emergency_blink_state), 0.0f, 0.0f);
		for(list<TrackChain3D *>::iterator i=emergencies.begin(); i!=emergencies.end(); ++i)
			(*i)->set_color(color);
		emergency_blink_timeout = t+0.5*Time::sec;
	}

	GL::Framebuffer::system().clear(GL::COLOR_BUFFER_BIT|GL::DEPTH_BUFFER_BIT);

	main_view.render();

	root->render();

	window.swap_buffers();
}

void Engineer::button_press(unsigned btn)
{
	if(btn==1)
	{
		Object *obj = pick_object(pointer);
		if(Track *track = dynamic_cast<Track *>(obj))
		{
			if(track->get_type().is_turnout())
			{
				Block &block = track->get_block();
				if(block.get_train() && block.get_train()->is_block_critical(block))
					set_status("Turnout is busy");
				else
				{
					unsigned paths = track->get_type().get_paths();
					unsigned i = track->get_active_path()+1;
					while(!(paths&(1<<i)))
					{
						if(!(paths>>i))
							i = 0;
						else
							++i;
					}
					track->set_active_path(i);
					set_status(format("Turnout %d", track->get_turnout_address()));
				}
			}
			if(unsigned saddr = track->get_sensor_address())
			{
				if(options.simulate)
					layout.get_driver().set_sensor(saddr, !layout.get_driver().get_sensor(saddr));
				set_status(format("Sensor %d", saddr));
			}
		}
		else if(Vehicle *veh = dynamic_cast<Vehicle *>(obj))
		{
			TrainDialog *dlg = new TrainDialog(*this, *veh->get_train());
			root->add(*dlg);
			dlg->autosize();
		}
	}
}

void Engineer::axis_motion(unsigned axis, float value, float)
{
	if(axis==0)
		pointer.x = value;
	if(axis==1)
		pointer.y = value;
}

void Engineer::view_all()
{
	const Layout3D::ObjectMap &objects = layout_3d.get_all();

	float view_aspect = float(window.get_width())/window.get_height();
	float view_height = tan(main_view.get_camera().get_field_of_view()/2.0f)*2.0f;
	float best_score = 0;
	GL::Vector3 pos;
	GL::Vector3 up;
	for(Angle angle; angle<Angle::half_turn(); angle+=Angle::from_radians(0.01))
	{
		Transform trans = Transform::rotation(-angle, Vector(0, 0, 1));
		BoundingBox bbox;
		for(Layout3D::ObjectMap::const_iterator i=objects.begin(); i!=objects.end(); ++i)
			bbox = bbox|trans.transform(i->second->get_object().get_bounding_box());

		const Vector &minp = bbox.get_minimum_point();
		const Vector &maxp = bbox.get_maximum_point();

		float width = maxp.x-minp.x;
		float height = maxp.y-minp.y;
		float aspect = width/height;
		float score = min(aspect/view_aspect, view_aspect/aspect);

		if(score>best_score)
		{
			best_score = score;

			float size = max(width/view_aspect, height);
			float c = cos(angle);
			float s = sin(angle);
			float x = (minp.x+maxp.x)/2;
			float y = (minp.y+maxp.y)/2;
			float z = max(size*1.05/view_height, 0.15);

			pos = GL::Vector3(c*x-s*y, s*x+c*y, z);
			up = GL::Vector3(-s, c, 0);
		}
	}

	GL::Camera &camera = main_view.get_camera();
	camera.set_position(pos);
	camera.set_up_direction(up);
	camera.set_look_direction(GL::Vector3(0, 0, -1));
	camera.set_aspect(float(window.get_width())/window.get_height());
	camera.set_depth_clip(pos.z*0.5, pos.z*1.5);
}

Object *Engineer::pick_object(const Vector &p)
{
	const GL::Vector3 &start = main_view.get_camera().get_position();
	GL::Vector4 ray = main_view.get_camera().unproject(GL::Vector4(p.x, p.y, 0, 0));

	// XXX Do this better; make this function a template?
	if(Vehicle *veh = layout.pick<Vehicle>(Ray(start, ray.slice<3>(0))))
		return veh;
	return layout.pick<Track>(Ray(start, ray.slice<3>(0)));
}

void Engineer::emergency(Block *block, const string &msg)
{
	set_status(msg);
	TrackChain3D *tch3d = new TrackChain3D(layout_3d, *block);
	tch3d->set_color(GL::Color(1.0f, 0.0f, 0.0f));
	tch3d->set_layer(2);
	emergencies.push_back(tch3d);
}

void Engineer::halt_event(bool h)
{
	if(!h)
	{
		for(list<TrackChain3D *>::iterator i=emergencies.begin(); i!=emergencies.end(); ++i)
			delete *i;
		emergencies.clear();
	}
}

void Engineer::locomotive_detected(const Driver::DetectedLocomotive &loco)
{
	if(!import_active)
	{
		NewTrainDialog *dlg = new NewTrainDialog(*this);
		dlg->prefill(loco);
		dlg->signal_response.connect(sigc::mem_fun(this, &Engineer::import_finished));
		root->add(*dlg);
		import_active = true;
	}
}

void Engineer::import_finished(int)
{
	import_active = false;
}

void Engineer::process_new_train(Train &train)
{
	Vehicle3D &loco3d = layout_3d.get_3d(train.get_vehicle(0));
	overlay->set_label(loco3d, train.get_name());
	train.signal_name_changed.connect(sigc::bind<0>(sigc::mem_fun(overlay, &Overlay3D::set_label), sigc::ref(loco3d)));
}

void Engineer::train_added(Train &train)
{
	new_trains.push_back(&train);

	GL::Color best_color;
	float best_d_sq = 0;
	for(unsigned i=0; i<10; ++i)
	{
		GL::Color color;
		unsigned h = rand()%3;
		color.r = (h==0 ? 0.0 : rand()*1.0/RAND_MAX);
		color.g = (h==1 ? 0.0 : rand()*1.0/RAND_MAX);
		color.b = (h==2 ? 0.0 : rand()*1.0/RAND_MAX);
		color = color*(1/max(max(color.r, color.g), color.b));
		float min_d_sq = 3;
		for(map<Train *, GL::Color>::const_iterator j=train_colors.begin(); j!=train_colors.end(); ++j)
		{
			float dr = color.r-j->second.r;
			float dg = color.g-j->second.g;
			float db = color.b-j->second.b;
			float d_sq = dr*dr+dg*dg+db*db;
			if(d_sq<min_d_sq)
				min_d_sq = d_sq;
		}
		if(min_d_sq>best_d_sq)
		{
			best_color = color;
			best_d_sq = min_d_sq;
		}
	}
	train_colors[&train] = best_color;

	Allocation3D *alloc = new Allocation3D(layout_3d, train);
	alloc->set_color(best_color);
}

void Engineer::sighandler(int sig)
{
	if(sig==SIGSEGV || sig==SIGILL || sig==SIGFPE || sig==SIGABRT)
	{
		signal(sig, SIG_DFL);
		IO::print(IO::cerr, "Fatal signal received, terminating\n");
		layout.get_driver().halt(true);
		layout.get_driver().flush();
		raise(sig);
	}
	else if(sig==SIGTERM || sig==SIGINT)
		exit(0);
}
