#ifndef LIBR2C2_ARDUCONTROL_H_
#define LIBR2C2_ARDUCONTROL_H_

#include <msp/core/mutex.h>
#include <msp/core/thread.h>
#include <msp/datafile/objectloader.h>
#include <msp/fs/path.h>
#include <msp/io/serial.h>
#include <msp/time/timedelta.h>
#include <msp/time/timestamp.h>
#include "driver.h"

namespace R2C2 {

class ArduControl: public Driver
{
public:
	class Loader: public Msp::DataFile::ObjectLoader<ArduControl>
	{
	public:
		Loader(ArduControl &);

	private:
		void mfx_announce_serial(unsigned);
		void mfx_locomotive(unsigned);
	};

private:
	enum Command
	{
		POWER_ON = 0x01,
		POWER_OFF = 0x02,
		READ_POWER_STATE = 0x03,
		READ_TRACK_CURRENT = 0x08,
		SET_OVERCURRENT_LIMIT = 0x09,
		READ_INPUT_VOLTAGE = 0x0A,
		MOTOROLA_SPEED = 0x11,
		MOTOROLA_REVERSE = 0x12,
		MOTOROLA_SPEED_DIRECTION = 0x13,
		MOTOROLA_SPEED_FUNCTION = 0x14,
		MOTOROLA_SOLENOID = 0x15,
		MFX_SET_STATION_ID = 0x21,
		MFX_ANNOUNCE = 0x22,
		MFX_SEARCH = 0x23,
		MFX_ASSIGN_ADDRESS = 0x24,
		MFX_PING = 0x25,
		MFX_SPEED = 0x28,
		MFX_SPEED_FUNCS8 = 0x29,
		MFX_SPEED_FUNCS16 = 0x2A,
		S88_READ = 0x30,
		SET_BAUD_RATE = 0x70,
		COMMAND_OK = 0x80,
		RECEIVE_OVERRUN = 0x81,
		FRAMING_ERROR = 0x82,
		INVALID_COMMAND = 0x83,
		LENGTH_ERROR = 0x84,
		INVALID_VALUE = 0x85,
		OVERCURRENT = 0xA0,
		BAUD_CHANGE_FAILED = 0xA1,
		TRACK_CURRENT = 0xC0,
		INPUT_VOLTAGE = 0xC1,
		POWER_STATE = 0xC2,
		S88_DATA = 0xD0,
		MFX_SEARCH_FEEDBACK = 0xD1,
		MFX_PING_FEEDBACK = 0xD2
	};

	struct Tag
	{
		enum Type
		{
			NONE,
			GENERAL,
			LOCOMOTIVE,
			ACCESSORY,
			SENSOR
		};

		Type type;
		unsigned char command;
		unsigned short serial;
		unsigned id;

		Tag();

		operator bool() const { return type!=NONE; }
	};

	enum GeneralCommand
	{
		POWER,
		NEW_LOCO
	};

	enum Protocol
	{
		MM,
		MFX
	};

	struct ProtocolInfo
	{
		unsigned max_address;
		unsigned max_speed;
		unsigned max_func;
	};

	template<typename T>
	struct ControlledVariable
	{
		T current;
		T pending;
		unsigned short serial;

		ControlledVariable(): current(), pending(), serial(0) { }
		ControlledVariable(T v): current(v), pending(v), serial(0) { }

		bool set(T v) { if(v==pending) return false; pending = v; ++serial; return true; }
		bool commit(unsigned short s) { if(s!=serial) return false; current = pending; return true; }
		void rollback() { pending = current; ++serial; }

		operator T() const { return current; }
	};

	struct Locomotive
	{
		enum Command
		{
			SPEED,
			REVERSE,
			FUNCTIONS
		};

		unsigned id;
		Protocol proto;
		unsigned address;
		ControlledVariable<unsigned> speed;
		ControlledVariable<bool> reverse;
		ControlledVariable<unsigned> funcs;
		unsigned last_change_age;

		Locomotive(Protocol, unsigned);

		unsigned create_speed_dir_command(char *) const;
		unsigned create_speed_func_command(unsigned, char *) const;
	};

	struct MfxInfo: public DetectedLocomotive
	{
		class Loader: public Msp::DataFile::ObjectLoader<MfxInfo>
		{
		public:
			Loader(MfxInfo &);
		};

		unsigned id;
	};

	struct Accessory
	{
		enum Kind
		{
			TURNOUT,
			SIGNAL
		};

		enum Command
		{
			ACTIVATE,
			DEACTIVATE
		};

		Kind kind;
		unsigned address;
		unsigned bits;
		unsigned valid_states;
		ControlledVariable<unsigned> state;
		unsigned uncertain;
		unsigned target;
		Msp::Time::TimeDelta active_time;

		Accessory(Kind, unsigned, unsigned, unsigned);

		unsigned create_state_command(unsigned, bool, char *) const;
	};

	struct Sensor
	{
		enum Command
		{
			STATE
		};

		unsigned address;
		ControlledVariable<bool> state;

		Sensor(unsigned);
	};

	struct PendingCommand
	{
		Tag tag;
		char command[15];
		unsigned char length;
		unsigned repeat_count;

		PendingCommand();
		PendingCommand(GeneralCommand);
		PendingCommand(Locomotive &, Locomotive::Command, unsigned = 0);
		PendingCommand(Accessory &, Accessory::Command, unsigned = 0);
	};

	template<typename T>
	class Queue
	{
	private:
		std::list<T> items;
		Msp::Mutex mutex;

	public:
		void push(const T &);
		bool pop(T &);
		bool empty() const;
	};

	class Task
	{
	protected:
		Task() { }
	public:
		virtual ~Task() { }

		virtual bool get_work(PendingCommand &) = 0;
		virtual void process_reply(const char *, unsigned) { }
	};

	class CommandQueueTask: public Task
	{
	private:
		Queue<PendingCommand> queue;

	public:
		virtual bool get_work(PendingCommand &);
	};

	class RefreshTask: public Task
	{
	private:
		typedef std::list<Locomotive *> LocomotivePtrList;

		LocomotivePtrList cycle;
		LocomotivePtrList::iterator next;
		unsigned round;
		Locomotive *loco;
		unsigned phase;
		Msp::Mutex mutex;

	public:
		RefreshTask();

		virtual bool get_work(PendingCommand &);

		void add_loco(Locomotive &);
		void remove_loco(Locomotive &);
	private:
		Locomotive *get_next_loco();
		void advance();
	};

	class S88Task: public Task
	{
	private:
		ArduControl &control;
		unsigned n_octets;
		unsigned octets_remaining;
		unsigned delay;

	public:
		S88Task(ArduControl &);

		virtual bool get_work(PendingCommand &);
		virtual void process_reply(const char *, unsigned);

		void set_n_octets(unsigned);
		void grow_n_octets(unsigned);
	};

	class MfxAnnounceTask: public Task
	{
	private:
		unsigned serial;
		Msp::Time::TimeStamp next;

	public:
		MfxAnnounceTask();

		virtual bool get_work(PendingCommand &);

		void set_serial(unsigned);
		unsigned get_serial() const { return serial; }
	};

	class MfxSearchTask: public Task
	{
	private:
		ArduControl &control;
		unsigned next_address;
		Msp::Time::TimeStamp next;
		unsigned size;
		unsigned bits;
		unsigned misses;
		Queue<MfxInfo> queue;

	public:
		MfxSearchTask(ArduControl &);

		virtual bool get_work(PendingCommand &);
		virtual void process_reply(const char *, unsigned);

		void set_next_address(unsigned);
		bool pop_info(MfxInfo &);
	};

	class MonitorTask: public Task
	{
	private:
		float voltage;
		float current;
		float base_level;
		float peak_level;
		Msp::Time::TimeStamp next_poll;
		unsigned next_type;

	public:
		MonitorTask();

		virtual bool get_work(PendingCommand &);
		virtual void process_reply(const char *, unsigned);

		float get_voltage() const { return voltage; }
		float get_current() const { return current; }
		void reset_peak();
		float get_peak() const { return peak_level-base_level; }
	};

	class ControlThread: public Msp::Thread
	{
	private:
		ArduControl &control;
		bool done;
		std::vector<Task *> tasks;

	public:
		ControlThread(ArduControl &);

		void exit();
	private:
		virtual void main();
		void init_baud_rate();
		bool get_work(PendingCommand &);
		unsigned do_command(const PendingCommand &, const Msp::Time::TimeDelta &);
		unsigned process_reply(const char *, unsigned);
		void set_power(bool);
	};

	typedef std::map<unsigned, Locomotive> LocomotiveMap;
	typedef std::vector<MfxInfo> MfxInfoArray;
	typedef std::map<unsigned, Accessory> AccessoryMap;
	typedef std::list<Accessory *> AccessoryPtrList;
	typedef std::map<unsigned, Sensor> SensorMap;

	Msp::IO::Serial serial;
	unsigned debug;
	Msp::FS::Path state_file;

	ControlledVariable<bool> power;
	bool halted;

	LocomotiveMap locomotives;
	MfxInfoArray mfx_info;
	AccessoryMap accessories;
	AccessoryPtrList accessory_queue;
	Accessory *active_accessory;
	unsigned char active_index;
	Msp::Time::TimeStamp off_timeout;

	SensorMap sensors;

	Msp::Time::TimeDelta command_timeout;
	Queue<PendingCommand> command_queue;
	Queue<Tag> completed_commands;
	RefreshTask refresh;
	S88Task s88;
	MfxAnnounceTask mfx_announce;
	MfxSearchTask mfx_search;
	MonitorTask monitor;
	ControlThread thread;

	static ProtocolInfo protocol_info[2];

public:
	ArduControl(const Options &);
	~ArduControl();

	virtual void set_power(bool);
	virtual bool get_power() const { return power; }
	virtual void halt(bool);
	virtual bool is_halted() const { return halted; }

	virtual const char *enumerate_protocols(unsigned) const;
private:
	static Protocol map_protocol(const std::string &);
public:
	virtual unsigned get_protocol_speed_steps(const std::string &) const;

	virtual const DetectedLocomotive *enumerate_detected_locos(unsigned) const;
	virtual unsigned add_loco(unsigned, const std::string &, const VehicleType &);
private:
	MfxInfoArray::iterator add_mfx_info(const MfxInfo &);
public:
	virtual void remove_loco(unsigned);
	virtual void set_loco_speed(unsigned, unsigned);
	virtual void set_loco_reverse(unsigned, bool);
	virtual void set_loco_function(unsigned, unsigned, bool);

	virtual unsigned add_turnout(unsigned, const TrackType &);
	virtual void remove_turnout(unsigned);
	virtual void set_turnout(unsigned, unsigned);
	virtual unsigned get_turnout(unsigned) const;

	virtual unsigned add_signal(unsigned, const SignalType &);
	virtual void remove_signal(unsigned);
	virtual void set_signal(unsigned, unsigned);
	virtual unsigned get_signal(unsigned) const;

private:
	unsigned add_accessory(Accessory::Kind, unsigned, unsigned, unsigned);
	void remove_accessory(Accessory::Kind, unsigned);
	void set_accessory(Accessory::Kind, unsigned, unsigned);
	unsigned get_accessory(Accessory::Kind, unsigned) const;
	void activate_accessory_by_mask(Accessory &, unsigned);

public:
	virtual unsigned add_sensor(unsigned);
	virtual void remove_sensor(unsigned);
	virtual void set_sensor(unsigned, bool) { }
	virtual bool get_sensor(unsigned) const;

	virtual void tick();
	virtual void flush();
private:
	void save_state() const;
};

} // namespace R2C2

#endif
