#include <queue>
#include <msp/core/maputils.h>
#include <msp/strings/format.h>
#include <msp/strings/utils.h>
#include "layout.h"
#include "route.h"
#include "track.h"
#include "trackiter.h"
#include "tracktype.h"
#include "zone.h"

using namespace std;
using namespace Msp;

namespace {

using namespace R2C2;

typedef std::pair<Track *, unsigned> Key;

struct Node
{
	TrackIter track;
	Node *prev;
	float dist;

	Node():
		prev(0), dist(0)
	{ }

	Node(const TrackIter &t):
		track(t), prev(0), dist(0)
	{ }

	Node(const TrackIter &t, Node &r, float d):
		track(t), prev(&r), dist(prev->dist+d)
	{ }

	bool operator<(const Node &other) const
	{ return dist>other.dist; }
};

struct TrackMatch
{
	Track &track;

	TrackMatch(Track &t): track(t) { }

	bool operator()(Track &t) const { return &t==&track; }
};

struct TrackInSet
{
	const set<Track *> &tracks;

	TrackInSet(const set<Track *> &t): tracks(t) { }

	bool operator()(Track &t) const { return tracks.count(&t); }
};

template<typename Pred>
list<Track *> dijkstra(const TrackIter &from, const Pred &goal)
{
	map<Key, Node> track_nodes;
	priority_queue<Node> nodes;
	Node *final = 0;

	nodes.push(from);

	while(!nodes.empty())
	{
		Node lowest = nodes.top();
		nodes.pop();

		Key key(lowest.track.track(), lowest.track.entry());
		if(track_nodes.count(key))
			continue;

		Node &ref = track_nodes[key] = lowest;
		if(goal(*lowest.track))
		{
			final = &ref;
			break;
		}

		const TrackType::Endpoint &ep = lowest.track.endpoint();
		for(unsigned i=0; ep.paths>>i; ++i)
			if(ep.has_path(i))
			{
				TrackIter next = lowest.track.next(i);
				if(!next)
					continue;

				if(track_nodes.count(Key(next.track(), next.entry())))
					continue;

				nodes.push(Node(next, ref, lowest.track->get_type().get_path_length(i)));
			}
	}

	list<Track *> result;
	for(Node *node=final; node; node=node->prev)
		result.push_front(&*node->track);

	return result;
}

template<typename Pred>
Route *create_route(const TrackIter &from, const Pred &goal)
{
	list<Track *> tracks = dijkstra(from, goal);

	if(tracks.empty())
		return 0;

	Route *route = new Route(from->get_layout());
	for(list<Track *>::iterator i=tracks.begin(); i!=tracks.end(); ++i)
		route->add_track(**i);

	route->set_name("Pathfinder");
	route->set_temporary(true);

	return route;
}

}


namespace R2C2 {

Route::Route(Layout &l):
	TrackChain(l),
	temporary(false)
{
	layout.add(*this);
}

Route::~Route()
{
	layout.remove(*this);
}

void Route::set_temporary(bool t)
{
	temporary = t;
}

void Route::update_turnouts()
{
	set<unsigned> found;
	for(set<Track *>::const_iterator i=tracks.begin(); i!=tracks.end(); ++i)
		if((*i)->get_type().is_turnout())
		{
			found.insert((*i)->get_turnout_address());
			update_turnout(**i);
		}

	// Remove any turnouts that do not exist in the route
	for(map<unsigned, int>::iterator i=turnouts.begin(); i!=turnouts.end();)
	{
		if(!found.count(i->first))
			turnouts.erase(i++);
		else
			++i;
	}
}

void Route::update_turnout(Track &track)
{
	if(!track.get_type().is_turnout())
		return;

	// Build a combined path mask from linked endpoints
	unsigned nls = track.get_n_link_slots();
	unsigned mask = track.get_type().get_paths();
	for(unsigned i=0; i<nls; ++i)
	{
		Track *link = track.get_link(i);
		if(!tracks.count(link))
			continue;

		unsigned ep_mask = track.get_type().get_endpoint(i).paths;
		if(link->get_type().is_turnout())
		{
			const TrackType::Endpoint &link_ep = link->get_type().get_endpoint(link->get_link_slot(track));
			int p = get_turnout(link->get_turnout_address());
			if(p>=0 && !link_ep.has_path(p))
				// The linked track is a turnout and has a path which is incompatible with this endpoint
				ep_mask = ~ep_mask;
		}

		mask &= ep_mask;
	}

	unsigned taddr = track.get_turnout_address();
	if(mask && !(mask&(mask-1)))
	{
		// Exactly one possible choice, set the path accordingly
		unsigned path = 0;
		for(; (mask && !(mask&1)); mask>>=1, ++path) ;
		turnouts[taddr] = path;
	}
	else if(!turnouts.count(taddr))
		// More than one possible choice, and no existing entry - set as undecided
		turnouts[taddr] = -1;
}

int Route::get_turnout(unsigned id) const
{
	map<unsigned, int>::const_iterator i = turnouts.find(id);
	if(i!=turnouts.end())
		return i->second;
	return -1;
}

unsigned Route::get_path(Track &trk) const
{
	if(trk.get_type().is_turnout())
	{
		map<unsigned, int>::const_iterator i = turnouts.find(trk.get_turnout_address());
		if(i!=turnouts.end() && i->second>=0)
			return i->second;
	}
	return trk.get_active_path();
}

void Route::add_turnout(Track &track, unsigned path)
{
	TrackSet tracks_to_add;
	tracks_to_add.insert(&track);
	unsigned nls = track.get_n_link_slots();
	for(unsigned i=0; i<nls; ++i)
		if(track.get_type().get_endpoint(i).has_path(path))
		{
			TrackIter iter = TrackIter(&track, i).flip();
			for(; (iter && !iter->get_type().is_turnout()); iter=iter.next())
				tracks_to_add.insert(&*iter);
		}

	add_tracks(tracks_to_add);
}

void Route::on_track_added(Track &track)
{
	update_turnout(track);
	unsigned nls = track.get_n_link_slots();
	for(unsigned i=0; i<nls; ++i)
		if(Track *link = track.get_link(i))
			if(tracks.count(link))
				update_turnout(*link);
}

void Route::save(list<DataFile::Statement> &st) const
{
	st.push_back((DataFile::Statement("name"), name));
	for(map<unsigned, int>::const_iterator i=turnouts.begin(); i!=turnouts.end(); ++i)
		st.push_back((DataFile::Statement("turnout"), i->first, i->second));
}

DataFile::Statement Route::save_reference() const
{
	return (DataFile::Statement("route"), name);
}

Route *Route::find(const TrackIter &from, Track &to)
{
	return create_route(from, TrackMatch(to));
}

Route *Route::find(const TrackIter &from, const Route &to)
{
	return create_route(from, TrackInSet(to.get_tracks()));
}

Route *Route::find(const TrackIter &from, const Zone &to)
{
	return create_route(from, TrackInSet(to.get_tracks()));
}

Route *Route::find(const TrackIter &from, const set<Track *> &to)
{
	return create_route(from, TrackInSet(to));
}


Route::Loader::Loader(Route &r):
	DataFile::ObjectLoader<Route>(r)
{
	add("name",    &Loader::name);
	add("turnout", &Loader::turnout);
}

void Route::Loader::finish()
{
	const set<Track *> &ltracks = obj.layout.get_all<Track>();
	map<Track *, unsigned> turnout_tracks;
	for(set<Track *>::const_iterator i=ltracks.begin(); i!=ltracks.end(); ++i)
		if((*i)->get_type().is_turnout())
		{
			TurnoutMap::iterator j = turnouts.find((*i)->get_turnout_address());
			if(j!=turnouts.end())
				turnout_tracks[*i] = j->second;
		}

	for(map<Track *, unsigned>::iterator i=turnout_tracks.begin(); i!=turnout_tracks.end(); )
	{
		if(obj.check_validity(*i->first)==VALID)
		{
			obj.add_turnout(*i->first, i->second);
			turnout_tracks.erase(i);
			i = turnout_tracks.begin();
		}
		else
			++i;
	}
}

void Route::Loader::name(const string &n)
{
	obj.name = n;
}

void Route::Loader::turnout(unsigned id, unsigned path)
{
	turnouts[id] = path;
}

} // namespace R2C2
