#ifndef LIBR2C2_TRAIN_H_
#define LIBR2C2_TRAIN_H_

#include <sigc++/signal.h>
#include <sigc++/trackable.h>
#include <msp/datafile/objectloader.h>
#include <msp/time/timedelta.h>
#include "blockallocator.h"
#include "controller.h"
#include "sensor.h"
#include "trainai.h"

namespace R2C2 {

class ArticleNumber;
class Block;
class SpeedQuantizer;
class Vehicle;
class VehicleType;

class Train: public sigc::trackable
{
public:
	class Loader: public Msp::DataFile::ObjectLoader<Train>
	{
	private:
		Block *prev_block;
		bool blocks_valid;

	public:
		Loader(Train &);
	private:
		virtual void finish();
		void blocks();
		void name(const std::string &);
		void quantized_speed();
		void router();
		void timetable();
		void vehicle(const std::string &);
	};

	sigc::signal<void, const std::string &> signal_name_changed;
	sigc::signal<void, const std::string &, float> signal_control_changed;
	sigc::signal<void, unsigned, bool> signal_function_changed;
	sigc::signal<void, TrainAI &, const TrainAI::Message &> signal_ai_event;
	sigc::signal<void, Block &> signal_advanced;
	sigc::signal<void, Block &> signal_rear_advanced;
	sigc::signal<void, unsigned, Vehicle &> signal_vehicle_added;
	sigc::signal<void, unsigned, Vehicle &> signal_vehicle_removed;

private:
	Layout &layout;
	const VehicleType &loco_type;
	unsigned address;
	std::string protocol;
	unsigned loco_id;
	std::string name;
	const Train *preceding_train;
	std::vector<Vehicle *> vehicles;
	BlockAllocator allocator;
	bool advancing;
	Controller *controller;
	std::list<TrainAI *> ais;
	unsigned current_speed_step;
	bool speed_changing;
	bool reverse;
	Msp::Time::TimeDelta stop_timeout;
	unsigned functions;

	BlockIter last_entry_block;
	Msp::Time::TimeDelta travel_time;
	bool pure_speed;
	SpeedQuantizer *speed_quantizer;
	bool accurate_position;
	float overshoot_dist;

public:
	Train(Layout &, const VehicleType &, unsigned, const std::string &);
	~Train();

	Layout &get_layout() const { return layout; }
	const VehicleType &get_locomotive_type() const { return loco_type; }
	unsigned get_address() const { return address; }
	const std::string &get_protocol() const { return protocol; }
	void set_name(const std::string &);
	const std::string &get_name() const { return name; }
	const Train *get_preceding_train() const { return preceding_train; }
	Controller &get_controller() const { return *controller; }

	void add_vehicle(const VehicleType &);
	void remove_vehicle(unsigned);
	unsigned get_n_vehicles() const;
	Vehicle &get_vehicle(unsigned);
	const Vehicle &get_vehicle(unsigned) const;

	void set_control(const std::string &, float);
	void set_function(unsigned, bool);
	float get_control(const std::string &) const;
	float get_speed() const;
	float get_quantized_speed() const;
	unsigned get_speed_step() const { return current_speed_step; }
	float get_maximum_speed() const;
	bool get_function(unsigned) const;
	unsigned get_functions() const { return functions; }

	void add_ai(TrainAI &);
	void remove_ai(TrainAI &);
	void ai_message(const TrainAI::Message &);

	template<typename T>
	T *get_ai_of_type() const
	{
		for(std::list<TrainAI *>::const_iterator i=ais.begin(); i!=ais.end(); ++i)
			if(T *ai = dynamic_cast<T *>(*i))
				return ai;
		return 0;
	}

	bool place(const BlockIter &);
	void unplace();
	bool is_placed() const { return !allocator.empty(); }
	void stop_at(Block *);
	bool is_block_critical(const Block &) const;
	BlockIter get_first_noncritical_block() const;
	void refresh_blocks_from(Block &);
	const BlockAllocator &get_block_allocator() const { return allocator; }
	float get_reserved_distance() const;

	void tick(const Msp::Time::TimeDelta &);

	void save(std::list<Msp::DataFile::Statement> &) const;
private:
	void control_changed(const Controller::Control &);
	void loco_speed_event(unsigned, unsigned, bool);
	void loco_func_event(unsigned, unsigned, bool);
	void advanced(Block &, Sensor *);
	void halt_event(bool);
	void block_reserved(const Block &, const Train *);
	float get_reserved_distance_until(const Block *) const;
};

} // namespace R2C2

#endif
